/*
 * Copyright (c) 2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 *  @brief Simple Zigbee network coordinator implementation
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/logging/log.h>
#include <dk_buttons_and_leds.h>

#include <zboss_api.h>
#include <zboss_api_addons.h>
#include <zb_mem_config_max.h>
#include <zigbee/zigbee_error_handler.h>
#include <zigbee/zigbee_app_utils.h>
#include <zb_nrf_platform.h>
#include "zb_range_extender.h"
#include <zb_zcl_door_lock.h>
#include <zb_zcl_reporting.h>


#define RUN_STATUS_LED                         DK_LED1
#define RUN_LED_BLINK_INTERVAL                 1000

/* Device endpoint, used to receive ZCL commands. */
#define ZIGBEE_COORDINATOR_ENDPOINT            10

/* Type of power sources available for the device.
 * For possible values see section 3.2.2.2.8 of ZCL specification.
 */
#define COORDINATOR_INIT_BASIC_POWER_SOURCE    ZB_ZCL_BASIC_POWER_SOURCE_DC_SOURCE

/* LED indicating that device successfully joined Zigbee network. */
#define ZIGBEE_NETWORK_STATE_LED               DK_LED3

/* LED used for device identification. */
#define IDENTIFY_LED                           DK_LED4

/* Button which reopens the Zigbee Network. */
#define KEY_ZIGBEE_NETWORK_REOPEN              DK_BTN1_MSK

/* Button used to enter the Identify mode. */
#define IDENTIFY_MODE_BUTTON                   DK_BTN4_MSK

/* If set to ZB_TRUE then device will not open the network after forming or reboot. */
#define ZIGBEE_MANUAL_STEERING                 ZB_FALSE

#define ZIGBEE_PERMIT_LEGACY_DEVICES           ZB_FALSE

#ifndef ZB_COORDINATOR_ROLE
#error Define ZB_COORDINATOR_ROLE to compile coordinator source code.
#endif

/* Button to start Factory Reset */
#define FACTORY_RESET_BUTTON IDENTIFY_MODE_BUTTON


#define MY_TEST_ENABLED

#ifdef MY_TEST_ENABLED
#define KEY_TEST_DEVICE_RESET              DK_BTN2_MSK
#define DOOR_LOCK_BUTTON                   DK_BTN3_MSK
#endif


LOG_MODULE_REGISTER(app, LOG_LEVEL_INF);


#ifdef MY_TEST_ENABLED
struct test_devive_context {
	zb_uint8_t endpoint;
	zb_uint16_t short_addr;
};

static struct test_devive_context test_ctx;


#endif

zb_uint8_t toggle;
/* Main application customizable context.
 * Stores all settings and static values.
 */
struct zb_device_ctx {
	zb_zcl_basic_attrs_t basic_attr;
	zb_zcl_identify_attrs_t identify_attr;
        zb_zcl_door_lock_attrs_t door_lock_attr;
};

/* Zigbee device application context storage. */
static struct zb_device_ctx dev_ctx;

ZB_ZCL_DECLARE_IDENTIFY_ATTRIB_LIST(
	identify_attr_list,
	&dev_ctx.identify_attr.identify_time);

ZB_ZCL_DECLARE_BASIC_ATTRIB_LIST(
	basic_attr_list,
	&dev_ctx.basic_attr.zcl_version,
	&dev_ctx.basic_attr.power_source);

ZB_ZCL_DECLARE_DOOR_LOCK_ATTRIB_LIST(
	door_lock_attr_list,
	&dev_ctx.door_lock_attr.lock_state,
	&dev_ctx.door_lock_attr.lock_type,
	&dev_ctx.door_lock_attr.actuator_enabled);

ZB_ZCL_DECLARE_IDENTIFY_CLIENT_ATTRIB_LIST(
	identify_client_attr_list);

ZB_DECLARE_RANGE_EXTENDER_CLUSTER_LIST(
	nwk_coordinator_clusters,
	basic_attr_list,
	identify_attr_list,
        door_lock_attr_list,
        identify_client_attr_list);

ZB_DECLARE_RANGE_EXTENDER_EP(
	nwk_coordinator_ep,
	ZIGBEE_COORDINATOR_ENDPOINT,
	nwk_coordinator_clusters);

ZBOSS_DECLARE_DEVICE_CTX_1_EP(
	nwk_coordinator,
	nwk_coordinator_ep);




#ifdef MY_TEST_ENABLED
/*
static void test_device_reset_to_factory_defaults(zb_bufid_t bufid)
{
	LOG_INF("Send reset to factory default command");

	ZB_ZCL_BASIC_SEND_RESET_REQ(bufid,
			       test_ctx.short_addr,
			       ZB_APS_ADDR_MODE_16_ENDP_PRESENT,
			       test_ctx.endpoint,
			       ZIGBEE_COORDINATOR_ENDPOINT,
			       ZB_AF_HA_PROFILE_ID,
			       ZB_ZCL_DISABLE_DEFAULT_RESPONSE,
			       NULL);
}


static void reset_to_factory_defaults_cmd(void)
{
	zb_buf_get_out_delayed(test_device_reset_to_factory_defaults);
}
*/

static void lock_door_cmd_cb(zb_bufid_t bufid)
{
    zb_zcl_door_lock_read_lock_door_res_payload_t *res = ZB_ZCL_DOOR_LOCK_READ_LOCK_DOOR_RES(bufid);
    if (res)
    {
        LOG_INF("Lock Door respond success, lock state: %d", res->status);
    }
    else
    {
        LOG_ERR("Lock Door respond fail");
    }
    zb_buf_free(bufid);
}

static void send_lock_door_cmd(zb_bufid_t bufid)
{
    if(toggle == 1) {
        toggle = 0;
        LOG_INF("Lock Door LOCK command send");
        ZB_ZCL_DOOR_LOCK_SEND_LOCK_DOOR_REQ(bufid,
                                        test_ctx.short_addr,
                                        ZB_APS_ADDR_MODE_16_ENDP_PRESENT,
                                        test_ctx.endpoint,  /* 서버의 엔드포인트 */
                                        ZIGBEE_COORDINATOR_ENDPOINT,
                                        ZB_ZCL_DISABLE_DEFAULT_RESPONSE,
                                        lock_door_cmd_cb);
    } else {
        toggle = 1;
         LOG_INF("Lock Door UNLOCK command send");
        ZB_ZCL_DOOR_LOCK_SEND_UNLOCK_DOOR_REQ(bufid,
                                        test_ctx.short_addr,
                                        ZB_APS_ADDR_MODE_16_ENDP_PRESENT,
                                        test_ctx.endpoint,  /* 서버의 엔드포인트 */
                                        ZIGBEE_COORDINATOR_ENDPOINT,
                                        ZB_ZCL_DISABLE_DEFAULT_RESPONSE,
                                        lock_door_cmd_cb);
    }
}


#endif


static void report_attr_cb(zb_zcl_addr_t *addr,
                           zb_uint8_t direction,
                           zb_uint16_t cluster_id,
                           zb_uint16_t attr_id,
                           zb_uint8_t attr_type,
                           zb_uint8_t *value)
{
    if (cluster_id == ZB_ZCL_CLUSTER_ID_DOOR_LOCK &&
        attr_id == ZB_ZCL_ATTR_DOOR_LOCK_LOCK_STATE_ID)
    {
        LOG_INF("[REPORT] Lock state changed! value = %d", *value);
    }
}



/**@brief Function for initializing all clusters attributes. */
static void app_clusters_attr_init(void)
{
	/* Basic cluster attributes data. */
	dev_ctx.basic_attr.zcl_version = ZB_ZCL_VERSION;
	dev_ctx.basic_attr.power_source = COORDINATOR_INIT_BASIC_POWER_SOURCE;

	/* Identify cluster attributes data. */
	dev_ctx.identify_attr.identify_time = ZB_ZCL_IDENTIFY_IDENTIFY_TIME_DEFAULT_VALUE;
}

/**@brief Function to toggle the identify LED.
 *
 * @param  bufid  Unused parameter, required by ZBOSS scheduler API.
 */
static void toggle_identify_led(zb_bufid_t bufid)
{
	static int blink_status;

	dk_set_led(IDENTIFY_LED, (++blink_status) % 2);
	ZB_SCHEDULE_APP_ALARM(toggle_identify_led, bufid, ZB_MILLISECONDS_TO_BEACON_INTERVAL(100));
}

/**@brief Function to handle identify notification events on the first endpoint.
 *
 * @param  bufid  Unused parameter, required by ZBOSS scheduler API.
 */
static void identify_cb(zb_bufid_t bufid)
{
	zb_ret_t zb_err_code;

	if (bufid) {
		/* Schedule a self-scheduling function that will toggle the LED. */
		ZB_SCHEDULE_APP_CALLBACK(toggle_identify_led, bufid);
	} else {
		/* Cancel the toggling function alarm and turn off LED. */
		zb_err_code = ZB_SCHEDULE_APP_ALARM_CANCEL(toggle_identify_led, ZB_ALARM_ANY_PARAM);
		ZVUNUSED(zb_err_code);

		dk_set_led(IDENTIFY_LED, 0);
	}
}

/**@brief Starts identifying the device.
 *
 * @param  bufid  Unused parameter, required by ZBOSS scheduler API.
 */
static void start_identifying(zb_bufid_t bufid)
{
	ZVUNUSED(bufid);

	if (ZB_JOINED()) {
		/* Check if endpoint is in identifying mode,
		 * if not, put desired endpoint in identifying mode.
		 */
		if (dev_ctx.identify_attr.identify_time ==
		    ZB_ZCL_IDENTIFY_IDENTIFY_TIME_DEFAULT_VALUE) {

			zb_ret_t zb_err_code = zb_bdb_finding_binding_target(
				ZIGBEE_COORDINATOR_ENDPOINT);

			if (zb_err_code == RET_OK) {
				LOG_INF("Enter identify mode");
			} else if (zb_err_code == RET_INVALID_STATE) {
				LOG_WRN("RET_INVALID_STATE - Cannot enter identify mode");
			} else {
				ZB_ERROR_CHECK(zb_err_code);
			}
		} else {
			LOG_INF("Cancel identify mode");
			zb_bdb_finding_binding_target_cancel();
		}
	} else {
		LOG_WRN("Device not in a network - cannot enter identify mode");
	}
}

/**@brief Callback used in order to visualise network steering period.
 *
 * @param[in]   param   Not used. Required by callback type definition.
 */
static void steering_finished(zb_uint8_t param)
{
	ARG_UNUSED(param);

	LOG_INF("Network steering finished");
	dk_set_led_off(ZIGBEE_NETWORK_STATE_LED);
}
static zb_bool_t binding_cb(short int source_addr,
                            zb_uint8_t *ieee_addr,
                            zb_uint8_t endpoint,
                            unsigned short cluster_id)
{
    LOG_INF("Find & Bind binding callback called:");
    LOG_INF("  Source short addr: 0x%04x", source_addr);
    LOG_INF("  Endpoint: %d, Cluster: 0x%04x", endpoint, cluster_id);
    return ZB_TRUE;
}
static void start_find_and_bind_initiator(zb_uint8_t param)
{
    LOG_INF("Starting Find & Bind (Initiator)");
    zb_bdb_finding_binding_initiator(ZIGBEE_COORDINATOR_ENDPOINT, binding_cb);
}
/**@brief Callback for button events.
 *
 * @param[in]   button_state  Bitmask containing buttons state.
 * @param[in]   has_changed   Bitmask containing buttons that has changed their state.
 */
static void button_changed(uint32_t button_state, uint32_t has_changed)
{
	/* Calculate bitmask of buttons that are pressed and have changed their state. */
	uint32_t buttons = button_state & has_changed;
	zb_bool_t comm_status;

	if (buttons & KEY_ZIGBEE_NETWORK_REOPEN) {
		(void)(ZB_SCHEDULE_APP_ALARM_CANCEL(steering_finished, ZB_ALARM_ANY_PARAM));

		comm_status = bdb_start_top_level_commissioning(ZB_BDB_NETWORK_STEERING);
		if (comm_status) {
			LOG_INF("Top level commissioning restated");
		} else {
			LOG_INF("Top level commissioning hasn't finished yet!");
		}
#ifdef MY_TEST_ENABLED
	} else if (buttons & KEY_TEST_DEVICE_RESET) {
		/*reset_to_factory_defaults_cmd();*/
                 ZB_SCHEDULE_APP_CALLBACK(start_find_and_bind_initiator, 0);
#endif
	}

	if (IDENTIFY_MODE_BUTTON & has_changed) {
		if (IDENTIFY_MODE_BUTTON & button_state) {
			/* Button changed its state to pressed */
		} else {
			/* Button changed its state to released */
			if (was_factory_reset_done()) {
				/* The long press was for Factory Reset */
				LOG_DBG("After Factory Reset - ignore button release");
			} else   {
				/* Button released before Factory Reset */

				/* Start identification mode */
				ZB_SCHEDULE_APP_CALLBACK(start_identifying, 0);
			}
		}
	}

        if (buttons & DOOR_LOCK_BUTTON) {
		LOG_INF("Button 3 Push");
		zb_buf_get_out_delayed(send_lock_door_cmd);
	}

	check_factory_reset_button(button_state, has_changed);
}

/**@brief Function for initializing LEDs and Buttons. */
static void configure_gpio(void)
{
	int err;

	err = dk_buttons_init(button_changed);
	if (err) {
		LOG_ERR("Cannot init buttons (err: %d)", err);
	}

	err = dk_leds_init();
	if (err) {
		LOG_ERR("Cannot init LEDs (err: %d)", err);
	}
}
/*static zb_bool_t binding_cb(short int source_addr,
                            zb_uint8_t *ieee_addr,
                            zb_uint8_t endpoint,
                            unsigned short cluster_id)
{
    LOG_INF("Find & Bind binding callback called:");
    LOG_INF("  Source short addr: 0x%04x", source_addr);
    LOG_INF("  Endpoint: %d, Cluster: 0x%04x", endpoint, cluster_id);
    return ZB_TRUE;
}
static void start_find_and_bind_initiator(zb_uint8_t param)
{
    LOG_INF("Starting Find & Bind (Initiator)");
    zb_bdb_finding_binding_initiator(ZIGBEE_COORDINATOR_ENDPOINT, binding_cb);
}*/

/**@brief Zigbee stack event handler.
 *
 * @param[in]   bufid   Reference to the Zigbee stack buffer used to pass signal.
 */
void zboss_signal_handler(zb_bufid_t bufid)
{
	/* Read signal description out of memory buffer. */
	zb_zdo_app_signal_hdr_t *sg_p = NULL;
	zb_zdo_app_signal_type_t sig = zb_get_app_signal(bufid, &sg_p);
	zb_ret_t status = ZB_GET_APP_SIGNAL_STATUS(bufid);
	zb_ret_t zb_err_code;
	zb_bool_t comm_status;
	zb_time_t timeout_bi;

	switch (sig) {
	case ZB_BDB_SIGNAL_DEVICE_REBOOT:
		/* BDB initialization completed after device reboot,
		 * use NVRAM contents during initialization.
		 * Device joined/rejoined and started.
		 */
		if (status == RET_OK) {
			if (ZIGBEE_MANUAL_STEERING == ZB_FALSE) {
				LOG_INF("Start network steering");
				comm_status = bdb_start_top_level_commissioning(
					ZB_BDB_NETWORK_STEERING);
				ZB_COMM_STATUS_CHECK(comm_status);
                                /*ZB_SCHEDULE_APP_CALLBACK(start_find_and_bind_initiator, 0);*/
			} else {
				LOG_INF("Coordinator restarted successfully");
			}
		} else {
			LOG_ERR("Failed to initialize Zigbee stack using NVRAM data (status: %d)",
				status);
		}
		break;

	case ZB_BDB_SIGNAL_STEERING:
		if (status == RET_OK) {
			if (ZIGBEE_PERMIT_LEGACY_DEVICES == ZB_TRUE) {
				LOG_INF("Allow pre-Zigbee 3.0 devices to join the network");
				zb_bdb_set_legacy_device_support(1);
			}

			/* Schedule an alarm to notify about the end of steering period.
			 */
			LOG_INF("Network steering started");
			zb_err_code = ZB_SCHEDULE_APP_ALARM(
				steering_finished, 0,
				ZB_TIME_ONE_SECOND *
				ZB_ZGP_DEFAULT_COMMISSIONING_WINDOW);
			ZB_ERROR_CHECK(zb_err_code);
		}
		break;

	case ZB_ZDO_SIGNAL_DEVICE_ANNCE: {
		zb_zdo_signal_device_annce_params_t *dev_annce_params =
			ZB_ZDO_SIGNAL_GET_PARAMS(
				sg_p, zb_zdo_signal_device_annce_params_t);

		LOG_INF("New device commissioned or rejoined (short: 0x%04hx)",
			dev_annce_params->device_short_addr);

#ifdef MY_TEST_ENABLED
		test_ctx.short_addr = dev_annce_params->device_short_addr;
		test_ctx.endpoint = 10;
#endif

		zb_err_code = ZB_SCHEDULE_APP_ALARM_CANCEL(steering_finished, ZB_ALARM_ANY_PARAM);
		if (zb_err_code == RET_OK) {
			LOG_INF("Joining period extended.");
			zb_err_code = ZB_SCHEDULE_APP_ALARM(
				steering_finished, 0,
				ZB_TIME_ONE_SECOND *
				ZB_ZGP_DEFAULT_COMMISSIONING_WINDOW);
			ZB_ERROR_CHECK(zb_err_code);
		}

	} break;

	default:
		/* Call default signal handler. */
		ZB_ERROR_CHECK(zigbee_default_signal_handler(bufid));
		break;
	}

	/* Update network status LED. */
	if (ZB_JOINED() &&
	    (ZB_SCHEDULE_GET_ALARM_TIME(steering_finished, ZB_ALARM_ANY_PARAM,
					&timeout_bi) == RET_OK)) {
		dk_set_led_on(ZIGBEE_NETWORK_STATE_LED);
	} else {
		dk_set_led_off(ZIGBEE_NETWORK_STATE_LED);
	}

	/*
	 * All callbacks should either reuse or free passed buffers.
	 * If bufid == 0, the buffer is invalid (not passed).
	 */
	if (bufid) {
		zb_buf_free(bufid);
	}
}


int main(void)
{
	int blink_status = 0;

	LOG_INF("Starting ZBOSS Coordinator example");

	/* Initialize */
	configure_gpio();
	register_factory_reset_button(FACTORY_RESET_BUTTON);

#ifdef MY_TEST_ENABLED
	test_ctx.short_addr = 0xFFFF;
#endif

	/* Register device context (endpoints). */
	ZB_AF_REGISTER_DEVICE_CTX(&nwk_coordinator);

	app_clusters_attr_init();

        /* 추가: 도어락 클러스터 client 초기화 */
         zb_zcl_door_lock_init_client();

        /* 리포트 수신 콜백 등록 */
        ZB_ZCL_SET_REPORT_ATTR_CB(report_attr_cb);

	/* Register handlers to identify notifications. */
	ZB_AF_SET_IDENTIFY_NOTIFICATION_HANDLER(ZIGBEE_COORDINATOR_ENDPOINT, identify_cb);

	/* Start Zigbee default thread */
	zigbee_enable();

	LOG_INF("ZBOSS Coordinator example started");

	while (1) {
		dk_set_led(RUN_STATUS_LED, (++blink_status) % 2);
		k_sleep(K_MSEC(RUN_LED_BLINK_INTERVAL));
	}
}
